from ..constitutive import *

from .steel import *


def choose(name,grade='default',type='linear'):
    dict=globals()[name]
    if type=='linear':
        E=dict[grade]['E']
        nv=dict[grade]['nv']
        material=LinearElastic(E,nv)
        material.name=f'{grade} {name}'
        return material


from .processor import PreProcessor,PostProcessor,PlotProcessor
from .solver import LinearSolver,OneStepSolver


class LinearFea(LinearSolver,PreProcessor,PostProcessor,PlotProcessor):
    """
    The simplest analysis, suits for linear material and only has one increment. 
    Firstly, set the boundary condition;
    Secondly, submit the task;
    Finally, view or plot the result.
    """
    def __init__(self, mesh):
        super().__init__(mesh)
    
    def submit(self,postprocess=True,nonlinear=False):
        """Before submit, set the boundary condition first."""
        self.preprocess()
        self.slove()
        if postprocess:
            self.forward()
            self.update_show_dict()


class OneStepFea(OneStepSolver,PreProcessor,PostProcessor,PlotProcessor):
    """
    The simplest analysis, suits for elastic material and only has one increment. 
    Firstly, set the boundary condition;
    Secondly, submit the task;
    Finally, view or plot the result.
    """
    def __init__(self, mesh):
        super().__init__(mesh)
    
    def submit(self,postprocess=True,nonlinear=False):
        """Before submit, set the boundary condition first."""
        self.preprocess()
        self.slove()
        if not self.solved: print('Unsuccessful!')
        if postprocess:
            self.forward()
            self.update_show_dict()


import numpy as np
import numpy.linalg as LA

from .tools import *


class CommonMaterial2D:
    """
    Common material includes vacuum calculating Jacobian automatically.
    """
    def __call__(self,strain,*variables,autograd=True):
        return self.forward(strain,*variables,autograd)

    def __init__(self):
        self.name='Unnamed material'
        self.tensorform=False
        self.plane_stress=False
        self.use_flatten=False
    
    def set_name(self,name):
        """Set name of the material"""
        self.name=name      

    def forward(self,strain,*variables,autograd=True):
        """
        Calculate stress by given strain and (optional) give the Jacobian of the increment.
        """
        self.strain=strain
        self.variables=variables
        self.stress=self.cal_stress(self.strain,*self.variables)
        if autograd: self.cal_Jacobian()
        return self.stress
    
    def cal_stress(self,strain,*variables):
        """
        Don't update self.stress and use member variables (self.strain and self.variables) in this function.
        """
        if self.tensorform:
            epsilon=self.__cal_strain_tensor(strain)
            sigma=self.cal_sigma(epsilon,*variables)
            return self.__cal_stress_vector(sigma)
        else:
            return np.zeros_like(strain)
    
    def cal_sigma(self,epsilon,*variables):
        """Calculate stress tensor (sigma) by strain tensor (epsilon)."""
        return np.zeros_like(epsilon)

    def __cal_strain_tensor(self,vector):
        return np.array([ 
            [vector[0],0.5*vector[2]],
            [0.5*vector[2],vector[1]]
        ])
    
    def __cal_stress_vector(self,tensor):
        return np.array([tensor[0,0],tensor[1,1],tensor[0,1]])
    
    def cal_Jacobian(self,diff=1e-3):
        self.Jacobian=np.zeros((3,3))
        for component in range(3):
            dstrain=np.zeros(3)
            dstrain[component]=diff
            strain_new = self.strain+dstrain
            stress_new=self.cal_stress(strain_new,*self.variables)
            dstress=stress_new-self.stress
            self.Jacobian[component]=dstress/diff


class LinearElastic(CommonMaterial2D):
    """
    Constitutive relationship of isotropic linear material.
    E: Young's modulus;
    nv: Poisson's ratio.
    """
    def __init__(self,E,nv):
        super().__init__()
        self.E=E 
        self.nv=nv 
        self.cal_D()

    def cal_D(self):
        E=self.E
        nv=self.nv
        if self.plane_stress: self.D=E/(1.0-nv**2)*\
            np.array([
            [1.0,nv,0.0], 
            [nv,1.0,0.0], 
            [0.0,0.0,0.5*(1.0-nv)]
        ])
        else: self.D=E/((1.0+nv)*(1.0-2.0*nv))*\
            np.array([
            [1.0-nv,nv,0.0], 
            [nv,1.0-nv,0.0], 
            [0.0,0.0,0.5*(1.0-2.0*nv)]
        ])
    
    def cal_stress(self,strain):
        return self.D@strain
    
    def cal_Jacobian(self):
        self.Jacobian=self.D


class TensorHookean(CommonMaterial2D):
    """
    mu: shear modulus;
    kappa: volum modulus.
    """
    def __init__(self,mu,kappa):
        super().__init__()
        self.mu=mu
        self.kappa=kappa
        self.tensorform=True
    
    @tensorflatten
    def cal_sigma(self, epsilon, *variables):
        trace=np.trace(epsilon)
        lamda=self.kappa-2.0/3.0*self.mu
        return 2.0*self.mu*epsilon+lamda*trace*np.eye(3)


class OtherMaterial(CommonMaterial2D):
    """
    Constitutive relationship of other material, you can define it by yourself.
    You should difine: 
    1. cal_stress(strain,*variables) or cal_sigma(epsilon,*variables) in tensor form;
    2. (optional) cal_Jacobian(strain,*variables).
    """
    pass


from multiprocessing import Pool, Process, process
import multiprocessing
from multiprocessing.shared_memory import SharedMemory
import numpy as np
import math

from . import kernel


class ParallelSupport(kernel.GlobalSystem):
    def __init__(self, mesh):
        super().__init__(mesh)
        self.processor_number=1
    
    def set_parallel(self,processor_number=None):
        if bool(processor_number): processor_number=multiprocessing.cpu_count()
        self.processor_number=processor_number

    def para_global_K(self):
        self.shm_K=SharedMemory(create=True,size=self.K.nbytes)
        self.K=np.ndarray(self.K.shape,self.K.dtype,self.shm_K.buf)
        self.K[:]=0.0
    
    def para_global_force(self):
        self.shm_force=SharedMemory(create=True,size=self.force.nbytes)
        self.force=np.ndarray(self.force.shape,self.force.dtype,self.shm_force.buf)
        self.force[:]=0.0
    
    def cut_elements(self):
        batch_size=math.ceil(len(self.elements)//self.processor_number)
        self.__batch_list=[batch*batch_size for batch in range(self.processor_number)]
        self.__batch_list.append(len(self.elements))
        start=self.__batch_list[:-1]
        end=self.__batch_list[1:]
        return map(range,start,end)
    
    def para_forward(self):
        self.para_global_force()
        self.para_global_K()
        element_indexes_list=self.cut_elements()
        with Pool(self.processor_number) as pool:
            pool.map(self.batch_forward,element_indexes_list)
        self.unlink_share_memory()

    def unlink_share_memory(self):
        self.K=self.K.copy()
        self.force=self.force.copy()
        self.shm_K.close()
        self.shm_force.close()
        self.shm_K.unlink()
        self.shm_force.unlink()

    def batch_forward(self,element_indexes):
        shm_K=SharedMemory(self.shm_K.name)
        self.K=np.ndarray(self.K.shape,self.K.dtype,shm_K.buf)
        shm_force=SharedMemory(self.shm_force.name)
        self.force=np.ndarray(self.force.shape,self.force.dtype,shm_force.buf)
        super().forward(element_indexes)

    def forward(self,element_indexes=None):
        if self.processor_number==1 or bool(element_indexes)!=None:
            super().forward(element_indexes)
        else:
            self.para_forward()


class LinearSolver(kernel.ReducedSystem,ParallelSupport):
    def __init__(self, mesh):
        super().__init__(mesh)
        self.solved=False

    def slove(self):
        self.backward()
        self.optimize()
        self.solved=True


class OneStepSolver(kernel.ReducedSystem,ParallelSupport):
    def __init__(self, mesh):
        super().__init__(mesh)
        self.simplest=False
        self.solved=False
        self.max_cycles=10
        self.tolerent_error=1e-8

    def slove(self):
        for _ in range(self.max_cycles):
            self.zero_global_K()
            self.zero_global_force()
            self.forward()
            self.backward()
            self.optimize()
            if self.error<self.tolerent_error: 
                self.solved=True
                break


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as tri


def plot_mesh(mesh):
    x=mesh.points.T[0]
    y=mesh.points.T[1]
    elements=mesh.cells_dict['triangle']
    mesh_fig=tri.Triangulation(x,y,elements)
    plt.figure()
    plt.gca().set_aspect('equal')
    plt.triplot(mesh_fig,'k.-',lw=1)
    plt.title('')

def cal_modulus(E='unkown',nv='unkown',K='unkown',G='unkown',lamda='unkown'):
    """You should and can only input two of elastic properties."""
    pass


import numpy as np


def flatten(cal_stress3d):
    def cal_stress2d(self,strain2d,*variables):
        strain3d=np.array([strain2d[0],strain2d[1],0.0,strain2d[2],0.0,0.0])
        stress3d=cal_stress3d(strain3d,*variables)
        return np.array([stress3d[0],stress3d[1],stress3d[3]])
    return cal_stress2d

def tensorflatten(cal_sigma3d):
    def cal_sigma2d(self,epsilon2d,*variables):
        epsilon3d=np.array([
            [epsilon2d[0,0],epsilon2d[0,1],0.0],
            [epsilon2d[1,0],epsilon2d[1,1],0.0],
            [0.0,0.0,0.0]
        ])
        sigma3d=cal_sigma3d(self,epsilon3d,*variables)
        return np.array([
            [sigma3d[0,0],sigma3d[0,1]],
            [sigma3d[1,0],sigma3d[1,1]]
        ])
    return cal_sigma2d


import numpy as np
from .. import element


class GlobalSystem:
    """
    The equation system is built for forward.
    """
    def __init__(self,mesh):
        self.simplest=True
        self.material_dict_location=dict()
        self.material_dict=dict()
        self.__get_mesh_data(mesh)

    def __get_mesh_data(self, mesh):
        self.nodes=mesh.points[:,:2]
        self.elements=mesh.cells_dict['triangle']

    # calculation for general material
    def init_global_system(self):
        self.init_global_vars()
        self.init_elements_vars()
        self.init_global_K()

    def init_global_vars(self):
        self.__len_global=2*len(self.nodes)
        self.zero_global_vars()
    
    def zero_global_vars(self):
        self.deform=np.zeros(self.__len_global)
        self.deform_obj=np.zeros(self.__len_global)
        self.force_obj=np.zeros(self.__len_global)
        self.ddeform=np.zeros(self.__len_global)
        self.force=np.zeros(self.__len_global)

    def zero_global_force(self):
        self.force=np.zeros(self.__len_global)

    def init_elements_vars(self):
        self.__len_elements=len(self.elements)
        self.zero_element_vars()

    def zero_element_vars(self):
        self.strain=np.zeros((self.__len_elements,3))
        self.stress=np.zeros((self.__len_elements,3))

    def init_global_K(self):
        self.__len_global=2*len(self.nodes)
        self.zero_global_K()
        if self.simplest: self.__cal_K()

    def zero_global_K(self):
        self.K=np.zeros((self.__len_global,self.__len_global))
    
    def __cal_K(self):
        for element_index,_ in enumerate(self.elements):
            element=self.__instant_Element(element_index)
            self.__Ke2K(element)
    
    def forward(self,element_indexes=None):
        if not bool(element_indexes): element_indexes=range(len(self.elements))
        for element_index in element_indexes:
            element=self.__instant_deformed_Element(element_index)
            self.__element_forward(element_index, element)
            self.__Fe2F(element)
            self.__Ke2K(element)
    
    def __instant_deformed_Element(self, element_index):
        element=self.__instant_Element(element_index)
        node_indexes=self.elements[element_index]
        deform_local=self.__get_deform_list(node_indexes)
        element.set_deform(deform_local)
        return element
    
    def __instant_Element(self,element_index):
        material=self.material_dict_location[element_index]
        node_indexes=self.elements[element_index]
        positions=np.array([self.nodes[node_indexes[j]][:] for j in range(3)])
        if self.simplest:
            element_instance=element.SimpleTriangleElement(material,positions)
        else:
            element_instance=element.TriangleElement(material,positions)
        element_instance.set_node_indexes(node_indexes)
        return element_instance
    
    def __get_deform_list(self,node_indexes):
        deform_local=[[self.deform[2*node],self.deform[2*node+1]] \
            for node in node_indexes.tolist()]
        return np.array(deform_local).reshape(-1)

    def __element_forward(self, element_index, element):
        element.forward()
        self.strain[element_index]=element.strain
        self.stress[element_index]=element.stress

    def __Fe2F(self,element):
        force_element=element.force
        node_indexes=element.node_indexes
        deform_global_index=self.__cal_element_map(node_indexes)
        for i_local,i_global in enumerate(deform_global_index):
            self.force[i_global]+=force_element[i_local]

    def __cal_element_map(self, node_indexes):
        return np.array([[2*node_indexes[i],2*node_indexes[i]+1] \
            for i in range(3)],dtype=np.uint64).reshape(-1)
    
    def __Ke2K(self,element):
        K_element=element.K_element
        node_indexes=element.node_indexes
        deform_global_index=self.__cal_element_map(node_indexes)
        for i_local,i_global in enumerate(deform_global_index):
            for j_local,j_global in enumerate(deform_global_index):
                self.K[i_global,j_global]+=K_element[i_local,j_local]


class ReducedSystem(GlobalSystem):
    """
    The equation system is built for backward.
    """
    def __init__(self, mesh):
        super().__init__(mesh)
    
    # built the reduced strain, stress, and stiffness
    def backward(self):
        self.__init_reduce_system()
        self.__solve_reduce_system()
    
    def __init_reduce_system(self):
        self.__init_reduce_vars()
        self.__update_reduce_force()
        self.__init_reduce_K()

    def __init_reduce_vars(self):
        self.__reduce_ddeform=np.zeros(self.len_reduce)
        self.__reduce_dforce=np.zeros(self.len_reduce)

    def __update_reduce_force(self):
        force=self.force_obj-self.K@self.deform_obj
        if not self.simplest: dforce=force-self.force
        if self.simplest:
            for i_reduce,i_global in enumerate(self.deform_free_index):
                self.__reduce_dforce[i_reduce]=force[i_global]
        else:
            for i_reduce,i_global in enumerate(self.deform_free_index):
                self.__reduce_dforce[i_reduce]=dforce[i_global]

    def __init_reduce_K(self):
        self.__reduce_K=np.zeros((self.len_reduce,self.len_reduce))
        self.__cal_reduce_K()

    def __cal_reduce_K(self):
        for i_reduce,i_global in enumerate(self.deform_free_index):
            for j_reduce,j_global in enumerate(self.deform_free_index):
                self.__reduce_K[i_reduce,j_reduce]=self.K[i_global,j_global]

    def __solve_reduce_system(self):
            self.__reduce_ddeform=np.linalg.solve(self.__reduce_K,self.__reduce_dforce)
            self.error=max(abs(self.__reduce_ddeform))

    def optimize(self):
        self.__update_ddeform()
        self.__update_global_deform()
        self.__update_global_force()

    def __update_ddeform(self):
        for i_reduce,i_global in enumerate(self.deform_free_index):
            self.ddeform[i_global]=self.__reduce_ddeform[i_reduce]

    def __update_global_deform(self):
        self.deform+=self.ddeform

    def __update_global_force(self):
            self.force+=self.K@self.ddeform




import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as tri

from . import kernel


class PreProcessor(kernel.ReducedSystem):
    def __init__(self, mesh):
        super().__init__(mesh)
        self.__init_condition()
        self.__init_show_dict()

    def __init_condition(self):
        self.x_given=set()
        self.y_given=set()
        self.x_given_displace=dict()
        self.y_given_displace=dict()
        self.f_given=dict()
    
    def __init_show_dict(self):
        self.__init_reference_dict()
        self.__init_deformed_dict()

    def __init_reference_dict(self):
        self.reference_dict={'position':{'x':[],'y':[]},'material':{'name':[],'color':[]}}
        self.reference_dict['position']['x']=self.nodes.T[0]
        self.reference_dict['position']['y']=self.nodes.T[1]

    def __init_deformed_dict(self):
        self.current_dict={'position':{'x':[],'y':[]},
            'deform':{'Ux':[],'Uy':[]},
            'force':{'Fx':[],'Fy':[]}, 
            'strain':{'e11':[],'e22':[],'e12':[]},
            'stress':{'S11':[],'S22':[],'S12':[]}}
    
    # preprocess
    #   set material
    def uniform_material(self,material,element_set='all'):
        """assign the given material to all elements of current analysis."""
        self.check_solved()
        if type(element_set)==str and element_set=='all':
            for element_index,_ in enumerate(self.elements):
                self.__update_material(material, element_index)
        else:
            for element_index in element_set:
                self.__update_material(material, element_index)

    def __update_material(self, material, element_index):
        if element_index > len(self.elements): raise ValueError
        if material not in self.material_dict_location.values() \
            and material.name in self.material_dict.values():
            material.name=material.name+'*'
        self.material_dict_location.update({element_index:material})
        self.material_dict.update({element_index:material.name})
    
    #   set boundary condition
    def set_deform_conditions(self,operation='fix',Ux=set(),Uy=set(),Uxy=set()):
        """assign deformation on the given nodes."""
        self.check_solved()
        if operation=='fix': 
            self.__fix_node_deform(Ux,Uy)
            self.__fix_node_deform(Uxy,Uxy)
        elif operation=='displace': 
            if type(Ux)==dict: self.__displace_for_x(Ux)
            if type(Uy)==dict: self.__displace_for_y(Uy)
            if type(Uxy)==dict: self.__displace_for_xy(Uxy)
        else:
            raise ValueError

    def __fix_node_deform(self, x_fix, y_fix):
        self.x_given.update(x_fix)
        self.y_given.update(y_fix)
        for x in x_fix:
            self.x_given_displace.update({x:0.0})
        for y in y_fix:
            self.y_given_displace.update({y:0.0})
    
    def __displace_for_x(self,x_displace):
        self.x_given_displace.update(x_displace)
        self.x_given.update(x_displace.keys())
    
    def __displace_for_y(self,y_displace):
        self.y_given_displace.update(y_displace)
        self.y_given.update(y_displace.keys())

    def __displace_for_xy(self,xy_condition):
        x_dict=dict()
        y_dict=dict()
        for node,displace in xy_condition.items():
            x_dict.update({node:displace[0]})
            y_dict.update({node:displace[1]})
        self.__displace_for_x(x_dict)
        self.__displace_for_y(y_dict)

    def set_force_conditions(self,f_given):
        """assign force load on the given nodes."""
        self.check_solved()
        self.f_given.update(f_given)
    
    def set_equation(self):
        pass
    
    def clear_conditions(self,*name):
        if 'Ux' in name:
            self.x_given=set()
            self.x_given_displace=dict()
        if 'Uy' in name:
            self.y_given=set()
            self.y_given_displace=dict()
        if 'Uxy' in name:
            self.x_given=set()
            self.x_given_displace=dict()
            self.y_given=set()
            self.y_given_displace=dict()
        if 'F' in name:
            self.f_given=dict()
        if 'all' in name or not bool(name):
            self.x_given=set()
            self.x_given_displace=dict()
            self.y_given=set()
            self.y_given_displace=dict()
            self.f_given=dict()
    
    def clear_node_conditions(self,node,*name):
        if 'Ux' in name:
            self.x_given.discard(node)
            self.x_given_displace.pop(node,0)
        if 'Uy' in name:
            self.y_given.discard(node)
            self.y_given_displace.pop(node,0)
        if 'Uxy' in name:
            self.x_given.discard(node)
            self.x_given_displace.pop(node,0)
            self.y_given.discard(node)
            self.y_given_displace.pop(node,0)
        if 'F' in name:
            self.f_given.discard(node)
        if 'all' in name or not bool(name):
            self.x_given.discard(node)
            self.x_given_displace.pop(node,0)
            self.y_given.discard(node)
            self.y_given_displace.pop(node,0)
            self.f_given.pop(node,0)

    def check_solved(self):
        if self.solved: 
            print('Conditions may have been changed! Please resubmit for new result.')
            self.sloved=False
    
    #   submit preprocess
    def preprocess(self):
        self.init_global_system()
        self.__init_reduce_map()
        self.__set_vars_obj()
        self.__fix_given_condition()

    def __init_reduce_map(self):
        self.deform_free_index=list()
        self.deform_fix_index=list()
        self.__cal_reduce_map()
        self.len_reduce=len(self.deform_free_index)
    
    def __cal_reduce_map(self):
        for node,_ in enumerate(self.nodes):
            if node not in self.x_given:
                self.deform_free_index.append(2*node)
            else:
                self.deform_fix_index.append(2*node)
            if node not in self.y_given:
                self.deform_free_index.append(2*node+1)
            else:
                self.deform_fix_index.append(2*node+1)
    
    def __set_vars_obj(self):
        self.deform_end=np.zeros_like(self.deform)
        self.force_end=np.zeros_like(self.force)
        for node,displace in self.x_given_displace.items():
            self.deform_end[2*node]=displace
        for node,displace in self.y_given_displace.items():
            self.deform_end[2*node+1]=displace
        for node,force in self.f_given.items():
            self.force_end[2*node],self.force_end[2*node+1]=force

    def __fix_given_condition(self):
        self.deform_obj=self.deform_end
        self.force_obj=self.force_end
        for i_global in self.deform_fix_index:
            self.deform[i_global]=self.deform_obj[i_global]


class PostProcessor(kernel.ReducedSystem):
    def __init__(self, mesh):
        super().__init__(mesh)
        self.__init_show_dict()
    
    def __init_show_dict(self):
        self.__init_reference_dict()
        self.__init_deformed_dict()

    def __init_reference_dict(self):
        self.reference_dict={'position':{'x':[],'y':[]},'material':{'name':[],'color':[]}}
        self.reference_dict['position']['x']=self.nodes.T[0]
        self.reference_dict['position']['y']=self.nodes.T[1]

    def __init_deformed_dict(self):
        self.current_dict={'position':{'x':[],'y':[]},
            'deform':{'Ux':[],'Uy':[]},
            'force':{'Fx':[],'Fy':[]}, 
            'strain':{'e11':[],'e22':[],'e12':[]},
            'stress':{'S11':[],'S22':[],'S12':[]}}
    
    #   data dict for viewing and ploting
    def update_show_dict(self):
        self.__update_show_deform()
        self.__update_show_position()
        self.__update_show_force()
        self.__update_show_elements_vars()

    def __update_show_deform(self):
        self.current_dict['deform']['Ux']=self.deform.reshape(len(self.nodes),2).T[0]
        self.current_dict['deform']['Uy']=self.deform.reshape(len(self.nodes),2).T[1]
    
    def __update_show_position(self):
        self.current_dict['position']['x']=self.reference_dict['position']['x']\
            +self.current_dict['deform']['Ux']
        self.current_dict['position']['y']=self.reference_dict['position']['y']\
            +self.current_dict['deform']['Uy']

    def __update_show_force(self):
        self.current_dict['force']['Fx']=self.force.reshape(len(self.nodes),2).T[0]
        self.current_dict['force']['Fy']=self.force.reshape(len(self.nodes),2).T[1]

    def __update_show_elements_vars(self):
        strain_list=np.array(self.strain).T
        stress_list=np.array(self.stress).T
        for i,name in enumerate(('e11','e22','e12')):
            self.current_dict['strain'][name]=strain_list[i]
        for i,name in enumerate(('S11','S22','S12')):
            self.current_dict['stress'][name]=stress_list[i]
    
    def get_data(self,name,index):
        """index can be node index or relement index."""
        dict=self.current_dict[name]
        return np.array([dict[key][index] for key in dict])
    
    def calculate(self,name):
        """calculate useful variables, including Mises stress, ..."""
        if name=='Mises':
            self.__cal_Mises()
        elif name=='Tresca':
            self.__cal_Tresca()
        else:
            raise ValueError

    def __cal_Mises(self):
        s11=self.current_dict['stress']['S11']
        s22=self.current_dict['stress']['S22']
        s12=self.current_dict['stress']['S12']
        Mises=np.sqrt(0.5*(s11**2+s22**2+(s11-s22)**2+6*s12**2))
        self.current_dict['stress']['Mises']=Mises

    def __cal_Tresca(self):
        s11=self.current_dict['stress']['S11']
        s22=self.current_dict['stress']['S22']
        s12=self.current_dict['stress']['S12']
        Tresca=np.sqrt((s11-s22)**2+4*s12**2)
        self.current_dict['stress']['Tresca']=Tresca


class PlotProcessor(kernel.ReducedSystem):
    def __init__(self, mesh):
        super().__init__(mesh)
        self.set_figsize()

    def set_figsize(self,figsize='small'):
        """figure size can be 'small', 'medium', 'large', 'verylarge', or (length,width)."""
        if figsize=='small':
            self.figsize=(4,4)
        elif figsize=='medium':
            self.figsize=(8,8)
        elif figsize=='large':
            self.figsize=(12,12)
        elif figsize=='verylarge':
            self.figsize=(16,16)
        else:
            self.figsize=figsize
    
    #   plot result figures
    def plot_mesh(self,node=True,element=False,deformed=False,magnification=None):
        if not deformed:
            self.__plot_undeformed_mesh(node,element)
        else:
            self.__plot_deformed_mesh(node,element,magnification)

    def __plot_undeformed_mesh(self,node=True,element=False):
        x=self.reference_dict['position']['x']
        y=self.reference_dict['position']['y']
        mesh_fig=tri.Triangulation(x,y,self.elements)
        plt.figure(figsize=self.figsize)
        plt.gca().set_aspect('equal')
        plt.triplot(mesh_fig,'k.-',lw=1)
        if node: self.__plot_node_index(x,y)
        if element: self.__plot_element_index(x,y)
        plt.title('Undeformed Mesh')
    
    def __plot_deformed_mesh(self,node=False,element=False,magnification=None):
        x,y,magnification=self.__amplify_deform(magnification)
        mesh_fig=tri.Triangulation(x,y,self.elements)
        plt.figure(figsize=self.figsize)
        plt.gca().set_aspect('equal')
        plt.triplot(mesh_fig,'k.-',lw=1)
        if node: self.__plot_node_index(x,y)
        if element: self.__plot_element_index(x,y)
        plt.title(f'Deformed Mesh \n(magnification = {magnification:.2e})')

    def __amplify_deform(self,magnification):
        if not bool(magnification): magnification=self.__cal_magnification()
        x=self.reference_dict['position']['x']+magnification*self.current_dict['deform']['Ux']
        y=self.reference_dict['position']['y']+magnification*self.current_dict['deform']['Uy']
        return x,y,magnification
    
    def __cal_magnification(self):
        max_position=max(max(abs(self.reference_dict['position']['x'])),\
            max(abs(self.reference_dict['position']['y'])))
        max_deform=max(max(abs(self.current_dict['deform']['Ux'])),\
            max(abs(self.current_dict['deform']['Uy'])))
        if max_deform>1e-10*max_position:
            magnification=0.1*max_position/max_deform
        else:
            magnification=0.0
        return magnification
    
    def __plot_node_index(self,x,y):
        for index,_ in enumerate(self.nodes):
            plt.annotate(index,(x[index],y[index]),color="red")
    
    def __plot_element_index(self,x,y):
        for index,element in enumerate(self.elements):
            x_mean=1/3*(x[element[0]]+x[element[1]]+x[element[2]])
            y_mean=1/3*(y[element[0]]+y[element[1]]+y[element[2]])
            plt.annotate(index,(x_mean,y_mean),xytext=(x_mean,y_mean))
    
    def plot_restrict(self,fix=True,load=True,node=False,element=False):
        x=self.reference_dict['position']['x']
        y=self.reference_dict['position']['y']
        mesh_fig=tri.Triangulation(x,y,self.elements)
        plt.figure(figsize=self.figsize)
        plt.gca().set_aspect('equal')
        plt.triplot(mesh_fig,'k.-',lw=1)
        if fix:
            if bool(self.x_given): self.__plot_x_fix()
            if bool(self.y_given): self.__plot_y_fix()
        if load:
            if bool(self.f_given): self.__plot_f_given()
        if node: self.__plot_node_index(x,y)
        if element: self.__plot_element_index(x,y)
        plt.title('Load and Restrict')
    
    def __plot_x_fix(self):
        x_fix_position=list()
        for node in self.x_given:
            x_fix_position.append(self.nodes[node])
        x_fix_position=np.array(x_fix_position).T
        x=x_fix_position[0]
        y=x_fix_position[1]
        plt.scatter(x,y,s=100,c='b',marker='>')
    
    def __plot_y_fix(self):
        y_fix_position=list()
        for node in self.y_given:
            y_fix_position.append(self.nodes[node])
        y_fix_position=np.array(y_fix_position).T
        x=y_fix_position[0]
        y=y_fix_position[1]
        plt.scatter(x,y,s=100,c='r',marker='^')
    
    def __plot_f_given(self):
        magnification=self.__cal_force_arrow()
        for node in self.f_given:
            x,y=self.nodes[node]
            dx=magnification*self.f_given[node][0]
            dy=magnification*self.f_given[node][1]
            head_width=0.3*max(abs(dx),abs(dy))
            plt.arrow(x,y,dx,dy,head_width=head_width,fc='g',ec='g')
    
    def __cal_force_arrow(self):
        max_position=max(max(abs(self.reference_dict['position']['x'])),\
            max(abs(self.reference_dict['position']['y'])))
        max_force=max([max(map(abs,value)) for value in self.f_given.values()])
        return 0.1*max_position/max_force
    
    def plot_material(self,node=False,element=False,deformed=False,magnification=None):
        self.__init_show_material()
        self.__plot_color('material','color','flat',deformed,magnification,colorbar=False)

    def __init_show_material(self):
        self.reference_dict['material']={'name':[],'color':[]}
        material_color=dict() #name:color
        for name in self.material_dict.values():
            self.reference_dict['material']['name'].append(name)
            if name not in material_color:
                material_color.update({name:len(material_color)})
            color=material_color[name]
            self.reference_dict['material']['color'].append(color)
    
    def plot(self,name,component='',node=False,element=False,deformed=True,magnification=None,colorbar=True):
        if name=='mesh':
            self.plot_mesh(node,element,deformed,magnification)
        elif name in ('deform','force'):
            self.__plot_color(name,component,'gouraud',deformed,magnification,colorbar)
        elif name in ('strain','stress'):
            self.__plot_color(name,component,'flat',deformed,magnification,colorbar)
        else:
            show_data=self.current_dict[name][component]
            if len(show_data)==len(self.nodes):
                self.__plot_color(name,component,'gouraud',deformed,magnification,colorbar)
            if len(show_data)==len(self.elements):
                self.__plot_color(name,component,'flat',deformed,magnification,colorbar)
    
    def __plot_color(self,name,component,shading,deformed,magnification,colorbar):
        if not deformed:
            x=self.reference_dict['position']['x']
            y=self.reference_dict['position']['y']
        else:
            x,y,magnification=self.__amplify_deform(magnification)
        if name in self.current_dict.keys():
            z=self.current_dict[name][component]
        if name in self.reference_dict.keys():
            z=self.reference_dict[name][component]
        fig=tri.Triangulation(x,y,self.elements)
        plt.figure(figsize=self.figsize)
        plt.gca().set_aspect('equal')
        plt.tripcolor(fig,z,shading=shading)
        plt.triplot(fig,lw=1)
        if colorbar: plt.colorbar()
        title=f'{name.capitalize()} ({component})'
        if deformed:  title+=f'\n(magnification = {magnification:.2e})'
        plt.title(title)


import numpy as np

class CommonElement:
    """plane triangle element"""
    def __init__(self, positions):
        """postion=ndarray([[x0,y0],[x1,y1],[x2,y2]])"""
        self.x=positions.T[0]
        self.y=positions.T[1]
        self.node_indexes=np.arange(len(positions),dtype=np.uint64)
    
    def set_deform(self,deform_element):
        self.deform_element=deform_element
    
    def set_node_indexes(self,node_indexes):
        self.node_indexes=node_indexes

class CommonTriangleElement(CommonElement):
    """plane triangle element"""
    def __init__(self, positions):
        super().__init__(positions)
        self.double_area=self.__cal_area_2()
        self.B=self.__cal_B()

    def __cal_area_2(self):
        x=self.x
        y=self.y
        matrix=np.array([
            [1.0,x[0],y[0]], 
            [1.0,x[1],y[1]], 
            [1.0,x[2],y[2]]])
        det=np.linalg.det(matrix)
        return abs(det)
    
    def __cal_B(self):
        x=self.x
        y=self.y
        return 1/self.double_area*np.array([
            [y[1]-y[2],0.0,y[2]-y[0],0.0,y[0]-y[1],0.0], 
            [0.0,x[2]-x[1],0.0,x[0]-x[2],0.0,x[1]-x[0]], 
            [x[2]-x[1],y[1]-y[2],x[0]-x[2],y[2]-y[0],x[1]-x[0],y[0]-y[1]]])

class TriangleElement(CommonTriangleElement):
    """plane triangle element for general elastic solid"""
    def __init__(self, material, positions):
        super().__init__(positions)
        self.material=material
    
    def forward(self):
        self.cal_strain()
        self.cal_stress(autograd=True)
        self.cal_Jacobian()
        self.cal_force()
    
    def cal_strain(self):
        self.strain=self.B@self.deform_element

    def cal_stress(self,autograd=False):
        self.material.forward(self.strain,autograd=autograd)
        self.stress=self.material.stress
    
    def cal_Jacobian(self):
        self.element_integrate()
    
    def element_integrate(self):
        self.K_element=0.5*self.double_area*self.B.T@self.material.Jacobian@self.B
    
    def cal_force(self):
        self.force=0.5*self.double_area*self.B.T@self.stress


class SimpleTriangleElement(TriangleElement):
    """plane triangle element for linear elastic solid"""
    def __init__(self, material, positions):
        super().__init__(material, positions)
        self.material=material
        self.element_integrate()
    
    def element_integrate(self):
        self.K_element=0.5*self.double_area*self.B.T@self.material.D@self.B
        